/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic                                  *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <glib/gprintf.h>

#include "engine.h"
#include "board.h"
#include "search.h"
#include "transpositionTable.h"

Board* board;
TranspositionTable* tTable;
GMainLoop* loop;
GIOChannel* uci_stdin;
guint timer = 0;

gboolean engineReadLine(GIOChannel* uci_stdin, GIOCondition condition, gpointer data);

#define numberOfNumberOptions 2
#define ttSize 64

/** Engine option representation. UI can set options to the engine.
 */
struct option{
	gint def;			/**< default option value */
	gint min;			/**< minimum option value */
	gint max;			/**< maximum option value */
	gchar name[255];	/**< option string name */
	void (*function)(); /**< function to call for setting of option to the engine */
};
typedef struct option Option;

void engineTTableReallocate(guint mSize){
	transpositionTableReallocate(tTable, mSize);
}

void engineSetMultiPv(guint mpv){
	searchSetMultiPV(mpv);
}

struct option numberOption[numberOfNumberOptions] = {
{ttSize,	1,	4096, 	"Hash", 	engineTTableReallocate},
{1,	1,	20,	   	"MultiPV",	engineSetMultiPv}
};

void engineRun(gchar book[]){
	boardInit(NULL,0);
	board = boardNew();
	tTable = transpositionTableNew(ttSize);
	loop = g_main_loop_new(NULL,FALSE);
	uci_stdin = g_io_channel_unix_new(0);
	if(!uci_stdin) g_error("Cannot create new uci GIOChannel!\n");
	if(!g_io_add_watch(uci_stdin,G_IO_IN, engineReadLine, NULL))
		g_error("Cannot add watch on GIOChannel!\n");
	if(!g_thread_supported ()) g_thread_init (0);
	else g_error("Threads not suported !!!\n");
	g_main_loop_run(loop);
}

inline void engineQuit(){
	boardFree(board);
	transpositionTableFree(tTable);
	g_io_channel_shutdown(uci_stdin,TRUE,NULL);
	g_main_loop_quit(loop);
	g_main_loop_unref(loop);
}

inline void engineIdentity(){
	guint i;
	g_printf("id name BBChess %0.2f\n",VERSION);
	g_printf("id author Borko Boskovic\n");
	for(i=0; i<numberOfNumberOptions; i++){
		g_printf("option name %s type spin default %i min %i max %i\n",
			numberOption[i].name, numberOption[i].def, numberOption[i].min, numberOption[i].max);
	}
	g_printf("uciok\n");
	fflush(stdout);
}

inline void engineIsReady(){
	g_printf("readyok\n");
	fflush(stdout);
}

void engineStopTimer(){
	if(timer != 0){
		if(!g_source_remove(timer)) g_error("Can not remove timer source!\n");
		timer = 0;
	}
}

gboolean engineStop(){
	engineStopTimer();
	searchStop();
	return FALSE;
}

void engineStartTimer(guint time){
	if(timer == 0) timer = g_timeout_add(time, engineStop, NULL);
	else g_error("Timer is alredy in use!\n");
}

inline guint engineTime(gint player, guint wtime, guint winc, guint btime, guint binc, guint movestogo){
	gint cwtime, cbtime;
	if(player == WHITE){
		if(movestogo == 1) return 0.8 * wtime;
		cwtime = wtime + (winc * (movestogo - 1));
		cbtime = btime + (binc * (movestogo - 1));
		cwtime = cwtime + (cwtime - cbtime)/2;
		cwtime = cwtime / movestogo;
		if(cwtime > 10000) cwtime = cwtime * 1.2;
		else cwtime = cwtime * 0.6;
		if(cwtime >= wtime || cwtime < 100) cwtime = wtime/2;
		return cwtime;
	}
	else{
		if(movestogo == 1) return 0.8 * btime;
		cwtime = wtime + (winc * (movestogo - 1));
		cbtime = btime + (binc * (movestogo - 1));
		cwtime = cwtime + (cwtime - cbtime)/2;
		cwtime = cwtime / movestogo;
		if(cwtime > 10000) cwtime = cwtime * 1.2;
		else cwtime = cwtime * 0.6;
		if(cwtime >= wtime || cwtime < 100) cwtime = wtime/2;	
		cwtime = cwtime + wtime + (winc * (movestogo - 1));
		cbtime = cbtime + (cbtime -cwtime)/2;
		cbtime = cbtime / movestogo;
		if(cbtime > 10000) cbtime = cbtime * 1.2;
		else cbtime = cbtime * 0.6;
		if(cbtime >= btime || cbtime < 100) cbtime = btime/2;
		return cbtime;
	}
}

inline void engineGo(gchar* line){
	searchReset();
	gchar command[20], *subline;
	guint value, wtime=0, winc=0, btime=0, binc=0, movestogo = 40;
	sscanf(line,"%s",command);
	searchSetDepth(40);
	if(!g_ascii_strcasecmp(command, "depth")){
		line = g_strstr_len(line, strlen(line), command) + strlen(command);
		sscanf(line,"%d",&value);
		searchSetDepth(value);
	}
	else if(!g_ascii_strcasecmp(command, "movetime")){
		line = g_strstr_len(line, strlen(line), command) + strlen(command);
		sscanf(line,"%d",&value);
		engineStartTimer(value);
	}
	else if(!g_ascii_strcasecmp(command, "infinite")){}
	else if(!g_ascii_strcasecmp(command, "mate")){}
	else if(!g_ascii_strcasecmp(command, "nodes")){
		line = g_strstr_len(line, strlen(line), command) + strlen(command);
		sscanf(line,"%d",&value);
		searchSetNodes(value);
	}
	else{
		subline = g_strstr_len(line, strlen(line), "wtime");
		if(subline != NULL) sscanf(subline,"wtime %d",&wtime);
		subline = g_strstr_len(line, strlen(line), "btime");
		if(subline != NULL) sscanf(subline,"btime %d",&btime);
		subline = g_strstr_len(line, strlen(line), "winc");
		if(subline != NULL) sscanf(subline,"winc %d",&winc);
		subline = g_strstr_len(line, strlen(line), "binc");
		if(subline != NULL) sscanf(subline,"binc %d",&binc);
		subline = g_strstr_len(line, strlen(line), "movestogo");
		if(subline != NULL) sscanf(subline,"movestogo %d",&movestogo);
		value = engineTime(board->player, wtime, winc, btime, binc, movestogo);
/*		g_printf("Time:%d\n",value);
		fflush(stdout);*/
		engineStartTimer(value);
	}
	searchStart(board,tTable);
}

inline void engineSetPosition(gchar* line){
	gchar command[20];
	sscanf(line,"%s",command);
	if(!g_ascii_strcasecmp(command, "startpos")){
		line = g_strstr_len(line, strlen(line), command) + strlen(command);
		sscanf(line,"%s",command);
		boardStartPosition(board);
		if(!g_ascii_strcasecmp(command, "moves")){
			line = g_strstr_len(line, strlen(line), command) + strlen(command);
			boardMakeMovesFromString(board,line);
		}
	}
	else if(!g_ascii_strcasecmp(command, "fen")){
		line = g_strstr_len(line, strlen(line), command) + strlen(command);
		boardSetFen(board,line);
		line = g_strstr_len(line, strlen(line), "moves");
		if(line != NULL){
			line = line + 6;
			boardMakeMovesFromString(board,line);
		}
	}
	#ifndef G_DISABLE_ASSERT
	boardPrint(board);
	gint i;
	Moves gmoves;
	boardGenerateMoves(board, &gmoves);
	for(i=0; i<gmoves.size; i++){
		if(!boardMakeMove(board,gmoves.move[i])){
					g_printf("Wrong move:");
					movePrint(gmoves.move[i]);
					g_printf("\n");
		}else{
					g_printf("Move:");
					movePrint(gmoves.move[i]);
					g_printf("\n");
					boardUnmakeMove(board);
		}
	}
	fflush(stdout);
	boardValidate(board);
	#endif
}

inline void engineSetOption(gchar* line){
	gchar command[20], name[255];
	gint i,value;
	sscanf(line,"%s",command);
	if(!g_ascii_strcasecmp(command, "name")){
		line = g_strstr_len(line, strlen(line), command) + strlen(command);
		command[0] = name[0] = '\0';
		do{
			if(name[0] != '\0') strcat(name," ");
			strcat(name,command);
			sscanf(line,"%s",command);
			line = g_strstr_len(line, strlen(line), command) + strlen(command);
		}while(g_ascii_strcasecmp(command,"value"));
		for(i=0; i<numberOfNumberOptions; i++){
			if(!g_ascii_strcasecmp(name,numberOption[i].name)){
				sscanf(line,"%d",&value);
				if(value >= numberOption[i].min && value <= numberOption[i].max)
					numberOption[i].function(value);
				else
					g_error("Wrong option value (%s:%d)\n",numberOption[i].name,value);
				break;
			}
		}
	}
	else{
		g_error("Wrong uci command!\n");
	}
}

inline void engineNewGame(){
}

gboolean engineReadLine(GIOChannel* uci_stdin, GIOCondition condition, gpointer data){
	gchar command[20];
	gchar* subline, *line;
	g_io_channel_read_line(uci_stdin, &line, NULL, NULL,NULL);
	sscanf(line,"%s",command);
	subline = g_strstr_len(line,strlen(line), command) + strlen(command);
	if(!g_ascii_strcasecmp(command, "stop")) engineStop();
	else if(!g_ascii_strcasecmp(command, "go")) engineGo(subline);
	else if(!g_ascii_strcasecmp(command, "position")) engineSetPosition(subline);
	else if(!g_ascii_strcasecmp(command, "uci")) engineIdentity();
	else if(!g_ascii_strcasecmp(command, "setoption")) engineSetOption(subline);
	else if(!g_ascii_strcasecmp(command, "ucinewgame"))	engineNewGame();
	else if(!g_ascii_strcasecmp(command, "isready")) engineIsReady();
	else if(!g_ascii_strcasecmp(command, "debug")) boardPrint(board);
	else if(!g_ascii_strcasecmp(command, "quit")) engineQuit();
	else { g_printf("Wrong uci command:%s\n",command); fflush(stdout); }
	g_free(line);
	return TRUE;
}
