/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic                                  *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it g_printf(/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <glib.h>
#include <math.h>
#include <glib/gprintf.h>

#include "search.h"
#include "board.h"
#include "engine.h"

gboolean searchStopCond;
guint maxDepth;
guint maxQDepth;
guint maxNodes;
guint searchDepth;
guint searchNodes;
GThread* searchThread = NULL;
gboolean (*searchStopConditions)(void);
gint historyHeuristic[64][64];
TranspositionTable* tTable = NULL;
guint multipv = 1;

gint quiescent(Board* board, gint alpha, gint beta, gint currentDepth){
	gint value;
	searchNodes ++;
	
	if(searchStopConditions()) return -EVAL_MATE;
	
	if(currentDepth > maxQDepth) maxQDepth = currentDepth;
	
	value = boardEvaluate(board, alpha, beta);
	if(value > alpha){
		if (value >= beta){
		       return value;
		}
		alpha = value;
	}
	Moves moves;
	boardGenerateQuiescentMoves(board, &moves);
	while(boardMakeNextMove(board,&moves)){
		value = -quiescent(board,-beta,-alpha,currentDepth+1);
		boardUnmakeMove(board);
		if(value > alpha){
			if(value >= beta){
			       return value;
			}
			alpha = value;
      	}
	}
	return alpha;
}

gint alphaBeta(Board* board, gint alpha, gint beta, gint depth, gboolean nullMove,
				gint currentDepth){
	gint value, tmpDepth, fmargin = 0;
	gboolean endgame = TRUE, extend = FALSE, fprune = FALSE;
	Move lastMove, bestMove=0, trMove = 0;
	guint fiftyMovesRule, repeat;
	TTData* td = NULL;
	
	searchNodes++;

	if(searchStopConditions()) return -EVAL_MATE;
	
	/* Draw */
	fiftyMovesRule = board50MovesRule(board);
	repeat = boardThreeTimeRepeat(board);
	
	if(repeat == 2 || fiftyMovesRule == 100){
		transpositionTableWrite(tTable,board->key,depth,0,EXACT,EMPTY,searchStopCond);
		return EVAL_DRAW;
	}
	
	/* Transposition table */
	td = transpositionTableFind(tTable, board->key);
	if(td != NULL && repeat == 0 && fiftyMovesRule<80){
		trMove = td->move;
		if(td->depth >= depth){
			if(td->flag == EXACT) return td->evaluate;
			else if(td->flag == ALPHA && td->evaluate > alpha){
				bestMove = td->move;
				alpha = td->evaluate;
				if(alpha >= beta) return alpha;
			}
			else if(td->flag == BETA && td->evaluate < beta){
				beta = td->evaluate;
				if(alpha >= beta) return beta;
			}
		}
	}
	
	/* Extend */
	if(boardIsInCheck(board)){
		depth ++;
		extend = TRUE;
	}
	
	/* Horizon */
	if(depth < 1){
	       return quiescent(board,alpha, beta, currentDepth+1);
	}

	/* Null move pruning */
	if (nullMove && depth > 2 && !extend && !boardIsInCheck(board) &&
		board->numberOfPieces[WHITE] > 3 && board->numberOfPieces[BLACK] > 3) {
		boardMakeNullMove(board);
		if(depth > 6) tmpDepth = depth - 4;
		else tmpDepth = depth - 3;
		value = -alphaBeta(board,-beta, -beta + 1, tmpDepth, FALSE,currentDepth+1);
		boardUnmakeNullMove(board);
		if(value >= beta){
			transpositionTableWrite(tTable,board->key,depth,value,ALPHA,EMPTY,searchStopCond);
			return value;
		}
	}

	/* Internal Iterative Deepening */
	tmpDepth = depth/2;
	if(trMove == 0 && depth > 3 && tmpDepth >= transpositionTableDepth(tTable,board->key)){
		alphaBeta(board,alpha, beta, tmpDepth, FALSE,currentDepth+1);
		td = transpositionTableFind(tTable, board->key);
		if(td != NULL) trMove = td->move;
	}

	/* Transposition table move */
	if(trMove != 0) {
		boardMakeMove(board,trMove);
		endgame = FALSE;
		value = -alphaBeta(board,-beta, -alpha, depth - 1, TRUE,currentDepth+1);
		lastMove = boardUnmakeMove(board);
		if(value > alpha) {
			if(value >= beta){
				transpositionTableWrite(tTable,board->key,depth,value,ALPHA,lastMove, searchStopCond);
				return value;
			}
			alpha = value;
			bestMove = lastMove;
		}
	}

	/* Futility pruning */
	if(depth <= 3 && !extend){
		value = board->material[board->player] - board->material[board->opponent];
		switch(depth){
			case 3:{ 
				if(value + 900 <= alpha){
					fprune = TRUE; 
					fmargin = 500;
					depth --;
					break;
				}	
			}
			case 2:{ 
				if(value + 500 <= alpha){
					fprune = TRUE;
					fmargin = 500;
					break; 
				}
			}
			case 1:{
				if(value + 200 <= alpha){
					fprune = TRUE; 
					fmargin = 200;
					break; 
				}
			}
		}
	}
	
	Moves moves;
	boardGenerateQuiescentMoves(board, &moves);
	while(boardMakeNextMove(board,&moves)){
		endgame = FALSE;
		if(!fprune || boardIsInCheck(board) || 
			board->material[board->opponent]-board->material[board->player]+fmargin > alpha)
		{
			value = -alphaBeta(board,-beta,-alpha , depth-1, TRUE,currentDepth+1);
			lastMove = boardUnmakeMove(board);
			if(value > alpha){
				if(value >= beta){
					transpositionTableWrite(tTable,board->key,depth,value,ALPHA,lastMove,searchStopCond);
					return value;
				}
				alpha = value;
				bestMove = lastMove;
			}
		}
		else boardUnmakeMove(board);
	}

	boardGenerateNonQuiescentMoves(board, &moves, historyHeuristic);
	while(boardMakeNextMove(board,&moves)){
		endgame = FALSE;
		if(!fprune ||  boardIsInCheck(board)){
			value = -alphaBeta(board,-beta,-alpha,depth-1,TRUE,currentDepth+1);
			lastMove = boardUnmakeMove(board);
			if(value > alpha){
				if(value >= beta){
					historyHeuristic[moveGetFrom(lastMove)][moveGetTo(lastMove)] += 1<<depth;
					transpositionTableWrite(tTable,board->key,depth,value,ALPHA,lastMove,searchStopCond);
					return value;
				}
				alpha = value;
				bestMove = lastMove;
			}
		}
		else boardUnmakeMove(board);
	}

	if(endgame){
		if(boardIsInCheck(board)){ /* Mate */
			transpositionTableWrite(tTable,board->key,depth,-EVAL_MATE,EXACT,EMPTY,searchStopCond);
			return -EVAL_MATE;
		}
		else{	/* Draw */
			transpositionTableWrite(tTable,board->key,depth,EVAL_DRAW,EXACT,EMPTY,searchStopCond);
			return EVAL_DRAW;
		}
	}
	
	if(bestMove != 0) transpositionTableWrite(tTable,board->key,depth,alpha,EXACT,bestMove,searchStopCond);
	else transpositionTableWrite(tTable,board->key,depth,alpha,BETA,EMPTY,searchStopCond);
	return alpha;
}

inline void printCurrentMove(guint index, Move move){
	g_printf("info currmovenumber %d currmove ",index);
	movePrint(move);
	g_printf("\n");
	fflush(stdout);
}

inline void printMultiPV(RootMoves* moves){
	guint i;
	if(moves->multipv == 1){
		g_printf("info ");
		g_printf("depth %d seldepth %d score cp %d time %.3f nodes %u nps %u hashfull %u pv ",
			moves->move[0].depth,
			moves->move[0].seldepth,
			moves->move[0].value, 
			moves->move[0].time, 
			moves->move[0].nodes,
			moves->move[0].nps,
			moves->move[0].hash);
		movePrint(moves->move[0].move);
		g_printf(" ");
		movePrintPV(moves->move[0].pv);
		g_printf("\n");
	}
	else{
		for(i=0; i<moves->multipv; i++){
			g_printf("info ");
			g_printf("multipv %d depth %d seldepth %d score cp %d time %.3f nodes %u nps %u hashfull %u pv ",
				i+1,
				moves->move[i].depth,
				moves->move[i].seldepth,
				moves->move[i].value, 
				moves->move[i].time, 
				moves->move[i].nodes,
				moves->move[i].nps,
				moves->move[i].hash);
			movePrint(moves->move[i].move);
			g_printf(" ");
			movePrintPV(moves->move[i].pv);
			g_printf("\n");
		}
	}
	fflush(stdout);
}

void alphaBetaRoot(Board* board, RootMoves* moves, gint depth, GTimer* timer){
	gint i, value, index, alpha, beta, window, current;
	guint nps;
	gdouble sec;
	gboolean up, down;
	RootMove move;
	maxQDepth = 0;
	searchNodes++;
		
	if(boardThreeTimeRepeat(board) == 2 || 
		board50MovesRule(board) == 100 || 
		moves->size == 0)
		g_error("Illegal position: can not make any move!\n");

	for(index=0; index<moves->size; index++){
		current = index;
		if(searchStopConditions()) break;
		if(depth > 9) printCurrentMove(current+1, moves->move[current].move);
		if(!boardMakeMove(board,moves->move[current].move)) g_error("Wrong move!\n");
		if(current < moves->multipv) value = moves->move[current].value;
		else value = moves->move[moves->multipv-1].value;
		window = 2;
		alpha = value - window/2;
		beta = value + window/2;
		up = down = FALSE;
		maxQDepth = 0;
		while(TRUE){
			if(current < moves->multipv){
				/*g_printf("ab(%d,%d)[%d]\n",alpha,beta,window);*/
				value = -alphaBeta(board,-beta,-alpha,depth-1,TRUE,0);
				if(searchStopConditions()) break;
				sec = g_timer_elapsed(timer,NULL);
				if(sec > 0) nps = (int)(searchNodes/sec);
				else nps = 0;
				moves->move[current].value = value;
				moves->move[current].depth = depth;
				moves->move[current].seldepth = maxQDepth;
				moves->move[current].time = sec;
				moves->move[current].nps = nps;
				moves->move[current].nodes = searchNodes;
				moves->move[current].hash = transpositionTableUsage(tTable);
				transpositionTableCollectingPV(tTable,board,moves->move[current].pv);
				for(i=0; i<current; i++){
					if(moves->move[i].value < value){
						memcpy(&move,&moves->move[current],sizeof(RootMove));
						while(i <= current){
							memcpy(&moves->move[current],&moves->move[current-1],sizeof(RootMove));
							current --;
						}
						memcpy(&moves->move[i],&move,sizeof(RootMove));
						current = i;
						break;
					}
				}
				printMultiPV(moves);
				if(value == EVAL_MATE) break;
				if(value >= beta){
					up = TRUE;
					if(up && down) break;
					window += window;
					alpha = value;
					beta = value + window;
					if(beta > EVAL_MATE) beta = EVAL_MATE+1;
					continue;
				}
				if(value <= alpha){
					down = TRUE;
					if(up && down) break;
					window += window;
					beta = value;
					alpha = value - window;
					if(alpha < -EVAL_MATE) alpha = -EVAL_MATE - 1;
					continue;
				}
			}
			else{
				alpha = value;
				beta = value + 1;
				value = -alphaBeta(board,-beta,-alpha,depth-1,TRUE,0);
				/*g_printf("mtd(%d,%d)=%d\n",alpha,beta,value);*/
				if(searchStopConditions()) break;
				moves->move[current].value = value;
				moves->move[current].depth = depth;
				transpositionTableCollectingPV(tTable,board,moves->move[current].pv);
				if(value <= alpha) break;
				else{
					for(i=0; i<moves->multipv; i++){
						if(moves->move[i].value < value){
							memcpy(&move,&moves->move[current],sizeof(RootMove));
							while(i <= current){
								memcpy(&moves->move[current],&moves->move[current-1],sizeof(RootMove));
								current --;
							}
							memcpy(&moves->move[i],&move,sizeof(RootMove));
							current = i;
							sec = g_timer_elapsed(timer,NULL);
							if(sec > 0) nps = (int)(searchNodes/sec);
							else nps = 0;
							moves->move[current].value = value;
							moves->move[current].depth = depth;
							moves->move[current].time = sec;
							moves->move[current].nps = nps;
							moves->move[current].nodes = searchNodes;
							moves->move[current].hash = transpositionTableUsage(tTable);
							printMultiPV(moves);
							break;
						}
					}
					up = TRUE;
					window = 4;
					alpha = value;
					beta = value + window;
					if(beta > EVAL_MATE) beta = EVAL_MATE+1;
					continue;
				}
			}
			break;
		}
		boardUnmakeMove(board);
	}
}

gpointer search(gpointer board){
	GTimer* timer = g_timer_new();
	guint nps;
	gdouble sec;
	gint i,j;
	for(i=0; i<64; i++)
		for(j=0; j<64; j++)
			historyHeuristic[i][j] = 0;
	#ifndef G_DISABLE_ASSERT
		boardEvaluate(board,0,0);
		return NULL;
	#endif
	transpositionTableReset(tTable);
	RootMoves moves;
	boardGenerateRootMoves(board, &moves);
	moves.multipv = MIN(moves.size,multipv);
	searchDepth = 1;
	if(maxDepth == 0){
		boardEvaluate(board, -EVAL_MATE,EVAL_MATE);
		g_printf("bestmove ");
		movePrint(moves.move[0].move);;
		g_printf("\n");
		fflush(stdout);
		return 0;
	}
	if(moves.size > 1){
		while (searchDepth <= maxDepth){
			g_printf("info depth %d\n",searchDepth);
			fflush(stdout);
			alphaBetaRoot((Board*)board, &moves, searchDepth, timer);
			if(searchStopConditions() || ABS(moves.move[0].value) == EVAL_MATE)
				break;
			searchDepth++;
			sec = g_timer_elapsed(timer,NULL);
			if(sec > 0) nps = (int)(searchNodes/sec);
			else nps = 0;
			g_printf("info time %.3f nodes %u nps %u hashfull %u\n",
					sec,searchNodes,nps,transpositionTableUsage(tTable));
			fflush(stdout);
		}
	}
	engineStopTimer();
	g_timer_destroy(timer);
	g_printf("bestmove ");
	movePrint(moves.move[0].move);;
	g_printf("\n");
	fflush(stdout);
	return NULL;
}

void searchStart(Board* board, TranspositionTable* trTable){
	tTable = trTable;
	if(searchThread != NULL) g_thread_join(searchThread);
	searchStopCond = FALSE;
	searchThread = g_thread_create(search,(gpointer)board,TRUE,NULL);
}

void searchStop(){
	searchStopCond = TRUE;
	if(searchThread != NULL){
		g_thread_join(searchThread);
		searchThread = NULL;
	}
}

gboolean searchStopCondition(void){
	return searchStopCond;
}

gboolean searchDepthCondition(void){
	if(searchStopCond || searchDepth > maxDepth){
		searchStopCond = TRUE;
		return TRUE;
	}
	else return FALSE;
}

gboolean searchNodesCondition(void){
	if(searchStopCond || searchNodes > maxNodes){
		searchStopCond = TRUE;
		return TRUE;
	}
	else return FALSE;
}

void searchReset(){
	maxDepth = maxNodes = searchDepth = searchNodes = 0;
	searchStopConditions =  searchStopCondition;
}

void searchSetDepth(guint depth){
	maxDepth = depth;
	searchStopConditions = searchDepthCondition;
}

void searchSetNodes(guint nodes){
	maxNodes = nodes;
	searchStopConditions = searchNodesCondition;
}

void searchSetMultiPV(guint nmultipv){
	multipv = nmultipv;
}
