/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic                                  *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
 * \file transpositionTable.h
 * This file contains functions for handling of the transposition table.
 */

#ifndef TRANSPOSITIONTABLE_H
#define TRANSPOSITIONTABLE_H

#include <glib.h>
#include "move.h"
#include "board.h"

#define EXACT 0 /**< Exact evaluation of position. */
#define ALPHA 1 /**< Alfa evaluation of position. */
#define BETA 2	/**< Beta evaluation of position. */

/**
 * Data entry of transposition table.
 */
struct ttData{
	guint64 key; 		/**< Zobrist hash key of the position. */
	gint8 depth;		/**< Depth of searched position. */
	gint8 flag;			/**< Type of evaluation. */
	gint16 evaluate;	/**< Evaluation of the position. */
	Move move;			/**< Selected move */
};
typedef struct ttData TTData; /**<  Type definition of transposition data. */

/**
 * Transposition table structure.
 */
struct transpositionTable{
	TTData* data; 	/**< Data of transposition table. */
	guint size;	/**< Size of transposition table. */
	guint free;	/**< Number of free data in transposition table. */
};
typedef struct transpositionTable TranspositionTable; /**< Type definition of transposition table */

/**
 * Creates new transposition table with a mSize megabytes of data.
 * \param mSize size of transposition table in MB.
 * \return pointer to newly created transposition table.
 */
TranspositionTable* transpositionTableNew(guint mSize);

/** Reallocates current transposition data to the new buffer with a
 * different size. 
 * \param tTable transposition table to reallocate
 * \param mSize new transposition size
 */
void transpositionTableReallocate(TranspositionTable* tTable, guint mSize);

/** Frees the given transposition table and its data.
 * \param tTable transposition table to consider
 */
void transpositionTableFree(TranspositionTable* tTable);

/**
 * Reports transposition table usage.
 * \param tTable table to consider
 */
guint transpositionTableUsage(TranspositionTable* tTable);

/** Writes data to the transposition table entry.
 * \param tTable transposition table to consider
 * \param key Zobrist hash key of the position
 * \param depth depth of search the position was encountered 
 * \param evaluate evaluation of the position
 * \param flag type of the evaluation (alpha, beta, exact)
 * \param move selected move 
 */
void transpositionTableWrite(TranspositionTable* tTable, guint64 key, gint depth, 
				guint evaluate, guint flag, Move move, gboolean searchStopCond);

/** Searches for a key in transposition table and returns the found
 * entry or NULL, if key is not found. 
 * \param tTable transposition table to consider
 * \param key Zobrist hash key of the position
 * \return transposition table entry if found, NULL otherwise
 */
TTData* transpositionTableFind(TranspositionTable* tTable, guint64 key);

void transpositionTableReset(TranspositionTable* tTable);

void transpositionTableCollectingPV(TranspositionTable* tTable, Board* board, Move pv[MAX_PV+1]);

gint transpositionTableDepth(TranspositionTable* tTable, guint64 key);

#endif
